from typing import Optional

from django.conf import settings
from django.db import transaction
from django_rq import job
from rq import Retry

from arkindex.process.models import ActivityState, Process, WorkerActivity, WorkerActivityState, WorkerVersion


@job("default", timeout=settings.RQ_TIMEOUTS["initialize_activity"], retry=Retry(max=4))
def initialize_activity(process: Process):
    """
    List all worker versions used in a process and initialize their activity on processed elements.
    4 retries allowed, for a total of 5 attempts, to try to mitigate some database errors from the large query.
    """
    try:
        with transaction.atomic():
            for version_id, configuration_id, model_version_id in process.worker_runs.values_list("version_id", "configuration_id", "model_version_id"):
                # Do not generate worker activities for the elements initialisation task
                if version_id != WorkerVersion.objects.init_elements_version.id:
                    WorkerActivity.objects.bulk_insert(
                        worker_version_id=version_id,
                        process_id=process.id,
                        configuration_id=configuration_id,
                        model_version_id=model_version_id,
                        elements_qs=process.list_elements()
                    )
    except Exception as e:
        process.activity_state = ActivityState.Error
        process.save()
        raise e
    process.activity_state = ActivityState.Ready
    process.save()


@job("default", timeout=settings.RQ_TIMEOUTS["process_delete"])
def process_delete(process: Process, delete_activity: Optional[WorkerActivityState] = None):
    """
    Delete a process and removes its reference from related worker activities.
    If `delete` is defined, activities of this state will be deleted.
    """
    if delete_activity:
        process.activities.filter(state=delete_activity).delete()
    process.delete()
