from datetime import datetime, timezone
from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.ponos.models import Farm
from arkindex.process.models import (
    CorpusWorkerVersion,
    FeatureUsage,
    GitRefType,
    ProcessMode,
    Repository,
    Revision,
    Worker,
    WorkerConfiguration,
    WorkerType,
    WorkerVersion,
    WorkerVersionState,
)
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Model, ModelVersionState
from arkindex.users.models import Role, User


class TestWorkersWorkerVersions(FixtureAPITestCase):
    """
    Test workers and worker versions endpoints
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.repo = Repository.objects.get(url="http://my_repo.fake/workers/worker")
        # User is the creator of the repository
        cls.repo.memberships.create(user=cls.user, level=Role.Admin.value)
        cls.rev = cls.repo.revisions.get()

        cls.worker_type_dla = WorkerType.objects.get(slug="dla")

        cls.worker_reco = Worker.objects.get(slug="reco")
        cls.worker_dla = Worker.objects.get(slug="dla")
        cls.worker_gpu = Worker.objects.get(slug="worker-gpu")
        cls.worker_generic = Worker.objects.get(slug="generic")
        cls.worker_file_import = Worker.objects.get(slug="file_import")
        cls.worker_custom = Worker.objects.get(slug="custom")
        cls.init_worker = Worker.objects.get(slug="initialisation")
        cls.worker_custom.memberships.create(user=cls.user, level=Role.Contributor.value)

        cls.rev2 = Revision.objects.create(
            hash="1234",
            message="commit message",
            author="Bob",
            repo=cls.repo,
        )

        cls.model = Model.objects.create(
            name="Generic model",
            public=False,
        )
        cls.model_version = cls.model.versions.create(
            state=ModelVersionState.Available,
            hash="42",
            archive_hash="42",
            size=1337,
        )

        cls.farm = Farm.objects.first()
        process = cls.corpus.processes.create(
            mode=ProcessMode.Workers,
            creator=cls.user,
            farm=cls.farm,
            name="test_process"
        )
        process.run()
        cls.task = process.tasks.get()

    def setUp(self):
        super().setUp()
        self.version_1 = WorkerVersion.objects.get(worker=self.worker_reco)
        self.version_1.gpu_usage = FeatureUsage.Disabled
        self.version_2 = self.worker_dla.versions.get()
        self.version_custom = self.worker_custom.versions.get()

    # Tests on get_query_set for WorkerList
    def test_workers_list_requires_login(self):
        response = self.client.get(reverse("api:workers-list"))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_workers_list(self):
        """
        User is able to list workers on the repository as they have an admin access
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:workers-list"))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 7,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(self.worker_custom.id),
                    "repository_id": None,
                    "name": "Custom worker",
                    "description": "",
                    "slug": "custom",
                    "type": "custom",
                    "archived": False,
                },
                {
                    "id": str(self.worker_dla.id),
                    "repository_id": str(self.repo.id),
                    "name": "Document layout analyser",
                    "description": "",
                    "slug": "dla",
                    "type": "dla",
                    "archived": False,
                },
                {
                    "id": str(self.init_worker.id),
                    "repository_id": None,
                    "name": "Elements Initialisation Worker",
                    "description": "",
                    "slug": "initialisation",
                    "type": "init",
                    "archived": False,
                },
                {
                    "id": str(self.worker_file_import.id),
                    "repository_id": str(self.repo.id),
                    "name": "File import",
                    "description": "",
                    "slug": "file_import",
                    "type": "import",
                    "archived": False,
                },
                {
                    "id": str(self.worker_generic.id),
                    "repository_id": str(self.repo.id),
                    "name": "Generic worker with a Model",
                    "description": "",
                    "slug": "generic",
                    "type": "recognizer",
                    "archived": False,
                },
                {
                    "id": str(self.worker_reco.id),
                    "repository_id": str(self.repo.id),
                    "name": "Recognizer",
                    "description": "",
                    "slug": "reco",
                    "type": "recognizer",
                    "archived": False,
                },
                {
                    "id": str(self.worker_gpu.id),
                    "repository_id": str(self.repo.id),
                    "name": "Worker requiring a GPU",
                    "description": "",
                    "slug": "worker-gpu",
                    "type": "worker",
                    "archived": False,
                },
            ]
        })

    def test_workers_list_compatible_model(self):
        """
        User is able to filter workers by a compatible model.
        """
        self.model.memberships.create(user=self.user, level=Role.Guest.value)
        self.worker_generic.models.set([self.model])
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:workers-list"),
                {"compatible_model": str(self.model.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(self.worker_generic.id),
                    "repository_id": str(self.repo.id),
                    "name": "Generic worker with a Model",
                    "description": "",
                    "slug": "generic",
                    "type": "recognizer",
                    "archived": False,
                },
            ]
        })

    def test_workers_list_compatible_model_public(self):
        public_model = Model.objects.create(name="Public", public=True)
        self.worker_reco.models.set([public_model])
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:workers-list"),
                {"compatible_model": str(public_model.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [{
                "id": str(self.worker_reco.id),
                "repository_id": str(self.repo.id),
                "name": "Recognizer",
                "description": "",
                "slug": "reco",
                "type": "recognizer",
                "archived": False,
            }],
        })

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_worker_list_compatible_model_private(self, filter_rights_mock):
        private_model = Model.objects.create(name="Private")
        self.worker_reco.models.set([private_model])
        self.client.force_login(self.user)
        filter_rights_mock.side_effect = [
            Worker.objects.all(),
            Repository.objects.all(),
            Model.objects.none(),
        ]

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:workers-list"),
                {"compatible_model": str(private_model.id)},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "compatible_model": ["This model does not exist or you don't have access to it"]
        })

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Worker, Role.Contributor.value),
            call(self.user, Repository, Role.Contributor.value),
            call(self.user, Model, Role.Guest.value),
        ])

    def test_workers_list_filter_invalid_query_params(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:workers-list"),
                {
                    "compatible_model": "A",
                    "repository_id": "A",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "compatible_model": ["Invalid UUID"],
            "repository_id": ["Invalid UUID"],
        })

    def test_workers_list_filter_type_slug(self):
        """
        User is able to filter workers on the repository by worker type slug
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:workers-list"), {"type": "dla"})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(self.worker_dla.id),
                    "repository_id": str(self.repo.id),
                    "name": "Document layout analyser",
                    "description": "",
                    "slug": "dla",
                    "type": "dla",
                    "archived": False,
                }
            ]
        })

    def test_workers_list_filter_archived(self):
        self.client.force_login(self.user)
        self.worker_dla.archived = datetime.now(timezone.utc)
        self.worker_dla.save()

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:workers-list"), {"archived": True})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(self.worker_dla.id),
                    "repository_id": str(self.repo.id),
                    "name": "Document layout analyser",
                    "description": "",
                    "slug": "dla",
                    "type": "dla",
                    "archived": True,
                }
            ]
        })

    def test_workers_list_filter_unarchived(self):
        self.client.force_login(self.user)
        Worker.objects.all().update(archived=datetime.now(timezone.utc))
        self.worker_dla.archived = None
        self.worker_dla.save()

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:workers-list"), {"archived": False})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(self.worker_dla.id),
                    "repository_id": str(self.repo.id),
                    "name": "Document layout analyser",
                    "description": "",
                    "slug": "dla",
                    "type": "dla",
                    "archived": False,
                }
            ]
        })

    def test_workers_list_filter_invalid_type_slug(self):
        """
        Raise when trying to filter workers with an invalid type slug
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:workers-list"), {"type": "invalid-slug"})
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"type": ["No registered worker type with that slug."]})

    def test_workers_list_filter_type_id(self):
        """
        User is able to filter workers on the repository by worker type id
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:workers-list"), {"type": self.worker_type_dla.id})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(self.worker_dla.id),
                    "repository_id": str(self.repo.id),
                    "name": "Document layout analyser",
                    "description": "",
                    "slug": "dla",
                    "type": "dla",
                    "archived": False,
                }
            ]
        })

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_workers_list_requires_contributor(self, filter_rights_mock):
        """
        User is not able to list workers with a guest access
        """
        self.client.force_login(self.user)
        filter_rights_mock.side_effect = [
            Worker.objects.none(),
            Repository.objects.none(),
        ]

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:workers-list"))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 0,
            "next": None,
            "number": 1,
            "previous": None,
            "results": []
        })

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Worker, Role.Contributor.value),
            call(self.user, Repository, Role.Contributor.value),
        ])

    def test_workers_list_name_filter(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:workers-list"), {"name": "layout"})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(self.worker_dla.id),
                    "repository_id": str(self.worker_dla.repository.id),
                    "name": "Document layout analyser",
                    "description": "",
                    "slug": "dla",
                    "type": "dla",
                    "archived": False,
                }
            ]
        })

    def test_workers_list_filter_repository(self):
        repo2 = Repository.objects.create(
            url="http://gitlab/repo2"
        )
        worker_2 = Worker.objects.create(
            repository=repo2,
            name="Worker 2",
            slug="worker_2",
            type=self.worker_type_dla
        )
        worker_2.memberships.create(user=self.user, level=Role.Contributor.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:workers-list"), {"repository_id": str(repo2.id)})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(worker_2.id),
                    "repository_id": str(repo2.id),
                    "name": "Worker 2",
                    "description": "",
                    "slug": "worker_2",
                    "type": self.worker_type_dla.slug,
                    "archived": False,
                }
            ]
        })

    def test_workers_retrieve_required_login(self):
        response = self.client.get(reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_workers_retrieve_repository(self):
        """
        A user can retrieve a worker by having contributor access on its repository
        """
        repo2 = Repository.objects.create(
            url="http://gitlab/repo2"
        )
        worker_2 = repo2.workers.create(name="Worker 2", slug="worker_2", type=self.worker_type_dla)
        repo2.memberships.create(user=self.user, level=Role.Contributor.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:worker-retrieve", kwargs={"pk": str(worker_2.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(worker_2.id),
            "repository_id": str(repo2.id),
            "name": "Worker 2",
            "description": "",
            "slug": "worker_2",
            "type": "dla",
            "archived": False,
        })

    def test_workers_retrieve_no_revision(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_custom.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.worker_custom.id),
            "repository_id": None,
            "name": "Custom worker",
            "description": "",
            "slug": "custom",
            "type": "custom",
            "archived": False,
        })

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_workers_retrieve_no_rights(self, filter_rights_mock):
        filter_rights_mock.side_effect = [
            Worker.objects.none(),
            Repository.objects.none(),
        ]
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(response.json(), {"detail": "No Worker matches the given query."})

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Worker, Role.Contributor.value),
            call(self.user, Repository, Role.Contributor.value),
        ])

    def test_worker_create_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse("api:workers-list"))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_worker_create_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:workers-list"))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_worker_create_required_fields(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:workers-list"))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "name": ["This field is required."],
            "slug": ["This field is required."],
            "type": ["This field is required."]
        })

    def test_worker_create_user(self):
        """
        A user can create a worker that is not linked to a repository.
        The user has an admin access to that worker.
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.post(
                reverse("api:workers-list"),
                data={
                    "name": "Worker post",
                    "slug": "worker_post",
                    "type": self.worker_type_dla.slug
                }
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        worker = Worker.objects.get(id=data["id"])
        self.assertEqual(worker.type, self.worker_type_dla)
        self.assertIsNone(worker.repository)
        self.assertQuerySetEqual(worker.memberships.values_list("user_id", "level"), [
            (self.user.id, Role.Admin.value)
        ])
        self.assertDictEqual(data, {
            "id": str(worker.id),
            "name": "Worker post",
            "description": "",
            "repository_id": None,
            "slug": "worker_post",
            "type": "dla",
            "archived": False,
        })

    @patch("arkindex.users.utils.get_max_level", return_value=Role.Guest.value)
    def test_worker_create_user_existing_no_execution_access(self, get_max_level_mock):
        """
        A user cannot retrieve a worker without an execution access in case the slug already exists.
        No worker type is created.
        """
        self.worker_custom.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:workers-list"),
                data={
                    "name": "Worker post",
                    "slug": self.worker_custom.slug,
                    "type": "new_type"
                }
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertFalse(WorkerType.objects.filter(slug="new_type").exists())
        self.assertQuerySetEqual(self.worker_custom.memberships.values_list("user_id", "level"), [
            (self.user.id, Role.Guest.value)
        ])
        self.assertDictEqual(response.json(), {"slug": ["You cannot use this value."]})
        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.worker_custom))

    def test_worker_create_task_return_existing_worker(self):
        """
        Creation with an existing slug returns the existing worker with a HTTP status 200
        """
        test_worker = Worker.objects.create(
            name="Prog Knife",
            slug="prog_knife",
            type=self.worker_reco.type
        )
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:workers-list"),
                data={
                    "name": "Progressive Knife",
                    "slug": "prog_knife",
                    "type": self.worker_type_dla.slug
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["id"], str(test_worker.id))
        self.assertEqual(data["name"], "Prog Knife")
        self.assertEqual(data["slug"], "prog_knife")
        self.assertEqual(data["type"], "recognizer")

    def test_worker_create_task_new_type(self):
        """
        Ponos task with an admin right on a repository can create a worker with a non-existing worker type
        """
        new_worker_type_slug = "newType"
        self.assertFalse(WorkerType.objects.filter(slug=new_worker_type_slug).exists())

        with self.assertNumQueries(11):
            response = self.client.post(
                reverse("api:workers-list"),
                data={
                    "name": "Worker post",
                    "slug": "worker_post",
                    "type": "newType"
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        self.assertTrue(WorkerType.objects.filter(slug=new_worker_type_slug).exists())
        self.assertNotEqual(data["id"], str(self.worker_reco.id))
        self.assertEqual(data["name"], "Worker post")
        self.assertEqual(data["slug"], "worker_post")
        self.assertEqual(data["type"], "newType")

    def test_update_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.put(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "name": "New name",
                    "slug": "new_slug",
                    "description": "New description",
                    "type": "dla",
                    "archived": False,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_update_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.put(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "name": "New name",
                    "slug": "new_slug",
                    "description": "New description",
                    "type": "dla",
                    "archived": False,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_update_requires_contributor(self, filter_rights_mock):
        self.worker_reco.memberships.update_or_create(user=self.user, defaults={"level": Role.Guest.value})
        self.repo.memberships.update_or_create(user=self.user, defaults={"level": Role.Guest.value})
        self.client.force_login(self.user)
        filter_rights_mock.side_effect = [
            Worker.objects.none(),
            Repository.objects.none(),
        ]

        with self.assertNumQueries(3):
            response = self.client.put(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "name": "New name",
                    "slug": "new_slug",
                    "description": "New description",
                    "type": "dla",
                    "archived": False,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(response.json(), {"detail": "No Worker matches the given query."})
        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Worker, Role.Contributor.value),
            call(self.user, Repository, Role.Contributor.value),
        ])

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_update(self, filter_rights_mock):
        self.client.force_login(self.user)
        filter_rights_mock.side_effect = [
            Worker.objects.all(),
            Repository.objects.none(),
        ]

        with self.assertNumQueries(6):
            response = self.client.put(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "name": "New name",
                    "slug": "new_slug",
                    "description": "New description",
                    "type": "dla",
                    "archived": False,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json(), {
            "id": str(self.worker_reco.id),
            "name": "New name",
            "slug": "new_slug",
            "description": "New description",
            "type": "dla",
            "repository_id": str(self.repo.id),
            "archived": False,
        })

        self.worker_reco.refresh_from_db()
        self.assertEqual(self.worker_reco.name, "New name")
        self.assertEqual(self.worker_reco.slug, "new_slug")
        self.assertEqual(self.worker_reco.description, "New description")
        self.assertEqual(self.worker_reco.type, self.worker_type_dla)
        self.assertEqual(self.worker_reco.repository_id, self.repo.id)

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Worker, Role.Contributor.value),
            call(self.user, Repository, Role.Contributor.value),
        ])

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_update_repository_contributor(self, filter_rights_mock):
        self.client.force_login(self.user)
        filter_rights_mock.side_effect = [
            Worker.objects.none(),
            Repository.objects.all(),
        ]

        with self.assertNumQueries(6):
            response = self.client.put(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "name": "New name",
                    "slug": "new_slug",
                    "description": "New description",
                    "type": "dla",
                    "archived": False,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json(), {
            "id": str(self.worker_reco.id),
            "name": "New name",
            "slug": "new_slug",
            "description": "New description",
            "type": "dla",
            "repository_id": str(self.repo.id),
            "archived": False,
        })

        self.worker_reco.refresh_from_db()
        self.assertEqual(self.worker_reco.name, "New name")
        self.assertEqual(self.worker_reco.slug, "new_slug")
        self.assertEqual(self.worker_reco.description, "New description")
        self.assertEqual(self.worker_reco.type, self.worker_type_dla)
        self.assertEqual(self.worker_reco.repository_id, self.repo.id)

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Worker, Role.Contributor.value),
            call(self.user, Repository, Role.Contributor.value),
        ])

    def test_update_unique_slug(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.put(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "name": "New name",
                    "slug": "dla",
                    "description": "New description",
                    "type": "dla",
                    "archived": False,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "non_field_errors": ["A worker with the same repository and slug already exists."],
        })

    @patch("arkindex.users.utils.get_max_level", return_value=Role.Contributor.value)
    def test_update_archived_requires_archivable(self, get_max_level_mock):
        self.client.force_login(self.user)
        self.assertFalse(self.worker_reco.is_archivable(self.user))

        cases = [
            (datetime.now(timezone.utc), False),
            (None, True),
        ]

        for current_value, new_value in cases:
            with self.subTest(current_value=current_value, new_value=new_value):
                get_max_level_mock.reset_mock()
                self.worker_reco.archived = current_value
                self.worker_reco.save()

                with self.assertNumQueries(4):
                    response = self.client.put(
                        reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                        {
                            "name": "New name",
                            "slug": "new_slug",
                            "description": "New description",
                            "type": "dla",
                            "archived": new_value,
                        },
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {
                    "archived": ["You are not allowed to archive or unarchive this worker."],
                })
                self.assertEqual(get_max_level_mock.call_count, 2)
                self.assertListEqual(get_max_level_mock.call_args_list, [call(self.user, self.worker_reco)] * 2)

    def test_update_archived(self):
        self.client.force_login(self.user)
        self.worker_reco.memberships.update_or_create(user=self.user, defaults={"level": Role.Admin.value})
        self.assertTrue(self.worker_reco.is_archivable(self.user))

        cases = [
            (datetime.now(timezone.utc), False),
            (None, True),
        ]

        for current_value, new_value in cases:
            with self.subTest(current_value=current_value, new_value=new_value):
                self.worker_reco.archived = current_value
                self.worker_reco.save()

                with self.assertNumQueries(6):
                    response = self.client.put(
                        reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                        {
                            "name": "New name",
                            "slug": "new_slug",
                            "description": "New description",
                            "type": "dla",
                            "archived": new_value,
                        },
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

                self.assertEqual(response.json(), {
                    "id": str(self.worker_reco.id),
                    "name": "New name",
                    "slug": "new_slug",
                    "description": "New description",
                    "type": "dla",
                    "repository_id": str(self.repo.id),
                    "archived": new_value,
                })

    def test_partial_update_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.patch(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "description": "New description",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_partial_update_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.patch(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "description": "New description",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_partial_update_requires_contributor(self, filter_rights_mock):
        self.worker_reco.memberships.update_or_create(user=self.user, defaults={"level": Role.Guest.value})
        self.repo.memberships.update_or_create(user=self.user, defaults={"level": Role.Guest.value})
        self.client.force_login(self.user)
        filter_rights_mock.side_effect = [
            Worker.objects.none(),
            Repository.objects.none(),
        ]

        with self.assertNumQueries(3):
            response = self.client.patch(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "description": "New description",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(response.json(), {"detail": "No Worker matches the given query."})

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Worker, Role.Contributor.value),
            call(self.user, Repository, Role.Contributor.value),
        ])

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_partial_update(self, filter_rights_mock):
        self.repo.memberships.filter(user=self.user).delete()
        self.worker_reco.memberships.update_or_create(user=self.user, defaults={"level": Role.Contributor.value})
        self.client.force_login(self.user)
        filter_rights_mock.side_effect = [
            Worker.objects.all(),
            Repository.objects.none(),
        ]

        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "description": "New description",
                    "type": "dla",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json(), {
            "id": str(self.worker_reco.id),
            "name": "Recognizer",
            "slug": "reco",
            "description": "New description",
            "type": "dla",
            "repository_id": str(self.repo.id),
            "archived": False,
        })

        self.worker_reco.refresh_from_db()
        self.assertEqual(self.worker_reco.name, "Recognizer")
        self.assertEqual(self.worker_reco.slug, "reco")
        self.assertEqual(self.worker_reco.description, "New description")
        self.assertEqual(self.worker_reco.type, self.worker_type_dla)
        self.assertEqual(self.worker_reco.repository_id, self.repo.id)

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Worker, Role.Contributor.value),
            call(self.user, Repository, Role.Contributor.value),
        ])

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_partial_update_repository_contributor(self, filter_rights_mock):
        self.worker_reco.memberships.filter(user=self.user).delete()
        self.repo.memberships.update_or_create(user=self.user, defaults={"level": Role.Contributor.value})
        self.client.force_login(self.user)
        filter_rights_mock.side_effect = [
            Worker.objects.none(),
            Repository.objects.all(),
        ]

        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "name": "New name",
                    "slug": "new_slug",
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json(), {
            "id": str(self.worker_reco.id),
            "name": "New name",
            "slug": "new_slug",
            "description": "",
            "type": "recognizer",
            "repository_id": str(self.repo.id),
            "archived": False,
        })

        self.worker_reco.refresh_from_db()
        self.assertEqual(self.worker_reco.name, "New name")
        self.assertEqual(self.worker_reco.slug, "new_slug")
        self.assertEqual(self.worker_reco.description, "")
        self.assertEqual(self.worker_reco.type.slug, "recognizer")
        self.assertEqual(self.worker_reco.repository_id, self.repo.id)

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Worker, Role.Contributor.value),
            call(self.user, Repository, Role.Contributor.value),
        ])

    def test_partial_update_unique_slug(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                {
                    "slug": "dla",
                },
            )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "non_field_errors": ["A worker with the same repository and slug already exists."],
        })

    @patch("arkindex.users.utils.get_max_level", return_value=Role.Contributor.value)
    def test_partial_update_archived_requires_archivable(self, get_max_level_mock):
        self.client.force_login(self.user)
        self.assertFalse(self.worker_reco.is_archivable(self.user))

        cases = [
            (datetime.now(timezone.utc), False),
            (None, True),
        ]

        for current_value, new_value in cases:
            with self.subTest(current_value=current_value, new_value=new_value):
                get_max_level_mock.reset_mock()
                self.worker_reco.archived = current_value
                self.worker_reco.save()

                with self.assertNumQueries(3):
                    response = self.client.patch(
                        reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                        {
                            "archived": new_value,
                        },
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {
                    "archived": ["You are not allowed to archive or unarchive this worker."],
                })
                self.assertEqual(get_max_level_mock.call_count, 2)
                self.assertListEqual(get_max_level_mock.call_args_list, [call(self.user, self.worker_reco)] * 2)

    def test_partial_update_archived(self):
        self.client.force_login(self.user)
        self.worker_reco.memberships.update_or_create(user=self.user, defaults={"level": Role.Admin.value})
        self.assertTrue(self.worker_reco.is_archivable(self.user))

        cases = [
            (datetime.now(timezone.utc), False),
            (None, True),
        ]

        for current_value, new_value in cases:
            with self.subTest(current_value=current_value, new_value=new_value):
                self.worker_reco.archived = current_value
                self.worker_reco.save()

                with self.assertNumQueries(4):
                    response = self.client.patch(
                        reverse("api:worker-retrieve", kwargs={"pk": str(self.worker_reco.id)}),
                        {
                            "archived": new_value,
                        },
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

                self.assertEqual(response.json(), {
                    "id": str(self.worker_reco.id),
                    "name": "Recognizer",
                    "slug": "reco",
                    "description": "",
                    "type": "recognizer",
                    "repository_id": str(self.repo.id),
                    "archived": new_value,
                })

    # Tests on get_query_set for WorkerVersionList
    def test_versions_list_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_versions_list(self):
        self.client.force_login(self.user)
        last_revision = self.repo.revisions.create(hash="42", message="A second version", author="Michel")
        last_version = self.worker_reco.versions.create(
            revision=last_revision,
            configuration={"a": "bc"},
            state=WorkerVersionState.Error,
        )
        last_version.created = "1999-09-09T09:09:09.090909Z"
        last_version.save()

        with self.assertNumQueries(9):
            response = self.client.get(reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["count"], 2)
        # The revision created in this test has a recent timestamp so it is displayed first
        first_version = data["results"][0]
        revision = first_version.pop("revision")
        self.assertEqual(revision["id"], str(last_revision.id))
        self.assertEqual(revision["message"], "A second version")
        self.assertDictEqual(first_version, {
            "id": str(last_version.id),
            "configuration": {"a": "bc"},
            "docker_image_iid": last_version.docker_image_iid,
            "state": "error",
            "gpu_usage": "disabled",
            "model_usage": FeatureUsage.Disabled.value,
            "worker": {
                "id": str(self.worker_reco.id),
                "name": self.worker_reco.name,
                "type": self.worker_reco.type.slug,
                "slug": self.worker_reco.slug,
            },
            "version": None,
            "created": "1999-09-09T09:09:09.090909Z",
        })

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_versions_list_requires_contributor(self, has_access_mock):
        """
        User is not able to list worker versions with a guest access
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {
            "detail": "You do not have an execution access to this worker.",
        })
        self.assertEqual(has_access_mock.call_count, 2)
        self.assertEqual(has_access_mock.call_args_list, [
            call(self.user, self.worker_reco, Role.Contributor.value, skip_public=False),
            call(self.user, self.repo, Role.Contributor.value, skip_public=False),
        ])

    def test_versions_list_public_worker(self):
        """
        Any authenticated user is able to list versions for public workers
        """
        user = User.objects.create_user("user42@test.fr", "abcd")
        user.verified_email = True
        user.save()
        self.worker_reco.public = True
        self.worker_reco.save()
        self.client.force_login(user)

        with self.assertNumQueries(9):
            response = self.client.get(reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["count"], 1)
        self.assertEqual(data["results"][0]["id"], str(self.worker_reco.versions.first().id))

    def test_versions_list_filter_worker(self):
        self.client.force_login(self.user)
        worker_2 = Worker.objects.create(
            repository=self.repo,
            name="Worker 2",
            slug="worker_2",
            type=self.worker_type_dla
        )
        version_2 = WorkerVersion.objects.create(
            worker=worker_2,
            revision=self.rev,
            configuration={"test": "test2"},
            gpu_usage=FeatureUsage.Disabled
        )

        with self.assertNumQueries(9):
            response = self.client.get(reverse("api:worker-versions", kwargs={"pk": str(worker_2.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(len(data["results"]), 1)
        version = data["results"][0]
        self.assertEqual(version["id"], str(version_2.id))

    def test_versions_list_simple_mode(self):
        """
        With the mode attribute set to simple, worker versions older than master/main
        or with no tag are excluded from the results
        """
        # An old branch revision
        non_tagged_rev = self.repo.revisions.create(hash="001", message="random", author="A1ice")
        non_tagged_rev.refs.create(name="a-second-branch", type=GitRefType.Branch, repository=self.repo)
        # An old tagged branch revision
        tagged_rev = self.repo.revisions.create(hash="002", message="v0.1", author="Bob")
        tagged_rev.refs.create(name="a-branch", type=GitRefType.Branch, repository=self.repo)
        tagged_rev.refs.create(name="0.1", type=GitRefType.Tag, repository=self.repo)
        # Master branch revision
        master_rev = self.repo.revisions.create(hash="003", message="up-to-date", author="B0b")
        master_rev.refs.create(name="master", type=GitRefType.Branch, repository=self.repo)
        # Main branch revision
        main_rev = self.repo.revisions.create(hash="004", message="up-to-date", author="B0b")
        main_rev.refs.create(name="main", type=GitRefType.Branch, repository=self.repo)
        # A recent revision with no reference
        recent_rev = self.repo.revisions.create(hash="005", message="up-to-date", author="AL1C3")
        WorkerVersion.objects.bulk_create([
            WorkerVersion(
                worker=self.worker_reco,
                revision=rev,
                configuration={"test": "42"}
            ) for rev in (non_tagged_rev, tagged_rev, master_rev, main_rev, recent_rev)
        ])

        # Complete mode
        with self.subTest(mode="complete"), self.assertNumQueries(8):
            response = self.client.get(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                {"mode": "complete"},
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.json()["count"], 6)

        # Simple mode
        with self.subTest(mode="simple"), self.assertNumQueries(8):
            response = self.client.get(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                {"mode": "simple"},
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            data = response.json()
            self.assertEqual(data["count"], 3)
            self.assertCountEqual(
                [version["revision"]["id"] for version in data["results"]],
                [str(tagged_rev.id), str(master_rev.id), str(main_rev.id)]
            )

    def test_create_version_non_existing_worker(self):
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": "12341234-1234-1234-1234-123412341234"}),
                data={"revision_id": str(self.rev2.id), "configuration": {"test": "test2"}},
                format="json",
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertDictEqual(response.json(), {"detail": "No Worker matches the given query."})

    def test_create_version_available_requires_docker_image(self):
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={
                    "revision_id": str(self.rev2.id),
                    "configuration": {"test": "test2"},
                    "state": "available",
                },
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "state": [
                "A worker version must have a `docker_image_iid` set when updated to available."
            ]
        })

    def test_create_version_return_existing_revision(self):
        # A worker version already exists for this worker and this revision
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={"revision_id": str(self.rev.id), "configuration": {"test": "test1"}},
                format="json",
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "revision_id": ["A version of this worker already exists with this revision"]
        })

    def test_create_version_empty(self):
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"configuration": ["This field is required."]})

    def test_create_version_archived(self):
        self.worker_reco.archived = datetime.now(timezone.utc)
        self.worker_reco.save()

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={
                    "revision_id": str(self.rev2.id),
                    "configuration": {},
                },
                format="json",
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "worker": ["This worker is archived."],
        })

    def test_create_version_unrelated_revision(self):
        """
        It is not possible to create a version between a worker and a revision of two different repositories
        """
        rev = Revision.objects.create(
            hash="1337",
            message="A worker import I am admin on",
            author="Yann",
            repo=Repository.objects.get(url="http://gitlab/repo")
        )

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={"revision_id": str(rev.id), "configuration": {"test": "test2"}},
                format="json",
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "revision_id": ["The revision must be part of the same repository as the worker."]
        })

    def test_create_version_revision_requires_worker_repository(self):
        self.worker_custom.memberships.filter(user=self.user).update(level=Role.Admin.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_custom.id)}),
                data={"revision_id": str(self.rev2.id), "configuration": {"test": "test2"}, "model_usage": FeatureUsage.Required.value},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "revision_id": ["A revision cannot be set on a worker that is not linked to a repository."]
        })

    def test_create_version_null_revision_requires_null_worker_repository(self):
        self.worker_custom.memberships.filter(user=self.user).update(level=Role.Admin.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={"configuration": {"test": "test2"}, "model_usage": FeatureUsage.Required.value},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "revision_id": ["A revision must be set on a worker that is linked to a repository."],
        })

    def test_create_version_null_worker_repository_requires_null_revision(self):
        self.worker_custom.memberships.filter(user=self.user).update(level=Role.Admin.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_custom.id)}),
                data={"revision_id": str(self.rev2.id), "configuration": {"test": "test2"}, "model_usage": FeatureUsage.Required.value},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "revision_id": ["A revision cannot be set on a worker that is not linked to a repository."]
        })

    def test_create_version_null_revision_forbidden_task_auth(self):
        """
        Ponos Task auth cannot create a version on a worker that is not linked to a repository.
        """
        self.worker_custom.versions.create(version=41, configuration={})

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_custom.id)}),
                data={"revision_id": str(self.rev2.id), "configuration": {"test": "test2"}, "model_usage": FeatureUsage.Required.value},
                format="json",
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "revision_id": ["Task authentication requires to create a version on workers linked to a repository."]
        })

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_create_version_user_auth_requires_admin(self, has_access_mock):
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_custom.id)}),
                data={"configuration": {"test": "val"}, "model_usage": FeatureUsage.Required.value},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You do not have an admin access to this worker."})

        self.assertEqual(has_access_mock.call_count, 1)
        self.assertEqual(has_access_mock.call_args, call(self.user, self.worker_custom, Role.Admin.value, skip_public=False))

    def test_create_version_user_auth_requires_null_repository(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
                data={"revision_id": str(self.rev2.id), "configuration": {"test": "test2"}, "model_usage": FeatureUsage.Required.value},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "revision_id": ["Ponos authentication is required to create a version on a worker linked to a repository."]
        })

    def test_create_version_null_revision(self):
        """
        A worker version can be created with no revision on a worker that has no repository.
        Its version number is automatically incremented.
        """
        self.worker_custom.memberships.filter(user=self.user).update(level=Role.Admin.value)
        self.worker_custom.versions.create(version=41, configuration={})
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_custom.id)}),
                data={"configuration": {"test": "val"}, "model_usage": FeatureUsage.Required.value},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        new_version = self.worker_custom.versions.get(id=data["id"])
        self.assertDictEqual(response.json(), {
            "id": str(new_version.id),
            "configuration": {"test": "val"},
            "created": new_version.created.isoformat().replace("+00:00", "Z"),
            "docker_image_iid": None,
            "gpu_usage": FeatureUsage.Disabled.value,
            "model_usage": FeatureUsage.Required.value,
            "revision": None,
            "state": WorkerVersionState.Created.value,
            "version": 42,
            "worker": {
                "id": str(self.worker_custom.id),
                "name": "Custom worker",
                "slug": "custom",
                "type": "custom",
            },
        })

    def test_create_version_configuration_wrong_type(self):
        """
        Configuration body must be an object
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                data={"configuration": "test"},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"configuration": ['Expected a dictionary of items but got type "str".']})

    def test_create_version(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
            data={"revision_id": str(self.rev2.id), "configuration": {"test": "test2"}, "model_usage": FeatureUsage.Required.value},
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        self.assertNotEqual(data["id"], str(self.version_1.id))
        self.assertEqual(data["configuration"], {"test": "test2"})
        self.assertEqual(data["revision"]["id"], str(self.rev2.id))
        self.assertEqual(data["state"], "created")
        self.assertEqual(data["gpu_usage"], "disabled")
        self.assertEqual(data["model_usage"], FeatureUsage.Required.value)

    def test_create_version_wrong_gpu_usage(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
            data={"revision_id": str(self.rev2.id), "configuration": {"test": "test2"}, "gpu_usage": "not_supported"},
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_version_no_user_configuration_ok(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {"beep": "boop"},
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_create_version_valid_user_configuration(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_integer": {"title": "Demo Integer", "type": "int", "required": True, "default": 1},
                        "demo_boolean": {"title": "Demo Boolean", "type": "bool", "required": False, "default": True},
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()["configuration"], {
            "user_configuration": {
                "demo_integer": {
                    "title": "Demo Integer",
                    "type": "int",
                    "required": True,
                    "default": 1
                },
                "demo_boolean": {
                    "title": "Demo Boolean",
                    "type": "bool",
                    "required": False,
                    "default": True
                }
            }
        })

    def test_create_version_valid_user_configuration_dict(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_dict": {"title": "Demo Dict", "type": "dict", "required": True, "default": {"a": "b", "c": "d"}},
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()["configuration"], {
            "user_configuration": {
                "demo_dict": {
                    "title": "Demo Dict",
                    "type": "dict",
                    "required": True,
                    "default": {"a": "b", "c": "d"}
                }
            }
        })

    def test_create_version_user_configuration_dict_strings_only(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_dict": {"title": "Demo Dict", "type": "dict", "required": True, "default": {"a": ["12", "13"], "c": "d"}},
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_version_valid_user_configuration_enum(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_choice": {"title": "Decisions", "type": "enum", "required": True, "default": 1, "choices": [1, 2, 3]}
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()["configuration"], {
            "user_configuration": {
                "demo_choice": {
                    "title": "Decisions",
                    "type": "enum",
                    "required": True,
                    "default": 1,
                    "choices": [1, 2, 3]
                }
            }
        })

    def test_create_version_valid_user_configuration_list(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_list": {"title": "Demo List", "type": "list", "required": True, "subtype": "int", "default": [1, 2, 3, 4]},
                        "boolean_list": {"title": "It's a list of booleans", "type": "list", "required": False, "subtype": "bool", "default": [True, False, False]}
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()["configuration"], {
            "user_configuration": {
                "demo_list": {
                    "title": "Demo List",
                    "type": "list",
                    "subtype": "int",
                    "required": True,
                    "default": [1, 2, 3, 4]
                },
                "boolean_list": {
                    "title": "It's a list of booleans",
                    "type": "list",
                    "subtype": "bool",
                    "required": False,
                    "default": [True, False, False]
                }
            }
        })

    def test_create_version_valid_user_configuration_model(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_model": {"title": "Model for training", "type": "model", "required": True},
                        "other_model": {"title": "Model the second", "type": "model", "default": str(self.model.id)}
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()["configuration"], {
            "user_configuration": {
                "demo_model": {
                    "title": "Model for training",
                    "type": "model",
                    "required": True
                },
                "other_model": {
                    "title": "Model the second",
                    "type": "model",
                    "default": str(self.model.id)
                }
            }
        })

    def test_create_version_invalid_user_configuration_list_requires_subtype(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_list": {"title": "Demo List", "type": "list", "required": True, "default": [1, 2, 3, 4]},
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "demo_list": {
                        "subtype": ['The "subtype" field must be set for "list" type properties.']
                    }
                }]
            }
        })

    def test_create_version_invalid_user_configuration_list_wrong_default(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_list": {"title": "Demo List", "type": "list", "required": True, "subtype": "int", "default": 12},
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "demo_list": {
                        "default": ["This is not a valid value for a field of type list."]
                    }
                }]
            }
        })

    def test_create_version_invalid_user_configuration_list_wrong_subtype(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_list": {"title": "Demo List", "type": "list", "required": True, "subtype": "dict", "default": [1, 2, 3, 4]},
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "demo_list": {
                        "subtype": ["Subtype can only be int, float, bool or string."]
                    }
                }]
            }
        })

    def test_create_version_invalid_user_configuration_list_wrong_default_subtype(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_list": {"title": "Demo List", "type": "list", "required": True, "subtype": "int", "default": [1, 2, "three", 4]},
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "demo_list": {
                        "default": ["All items in the default value must be of type int."]
                    }
                }]
            }
        })

    def test_create_version_invalid_user_configuration_not_list_choices(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_choice": {"title": "Decisions", "type": "enum", "required": True, "default": 1, "choices": "eeee"}
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "demo_choice": {
                        "choices": ['Expected a list of items but got type "str".']
                    }
                }]
            }
        })

    def test_create_version_invalid_user_configuration_not_dict(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": "non"
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"configuration": {"user_configuration": [['Expected a dictionary of items but got type "str".']]}})

    def test_create_version_invalid_user_configuration_item_not_dict(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "secrets": ["aaaaaa"]
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"configuration": {
            "user_configuration": [{
                "secrets":
                {"__all__": ["User configuration field definitions should be of type dict, not list."]}
            }]
        }})

    def test_create_version_invalid_user_configuration_wrong_field_type(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "something": {
                            "title": "some thing",
                            "type": "uh oh",
                            "required": 2
                        }
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "something": {
                        "required": ["Must be a valid boolean."],
                        "type": ["Value is not of type UserConfigurationFieldType"]
                    }
                }]
            }
        })

    def test_create_version_invalid_user_configuration_wrong_default_type(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "one_float": {
                            "title": "a float",
                            "type": "float",
                            "default": "bonjour",
                            "required": True
                        }
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "one_float": {
                        "default": ["This is not a valid value for a field of type float."]
                    }
                }]
            }
        })

    def test_create_version_invalid_user_configuration_choices_no_enum(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "something": {
                            "title": "some thing",
                            "type": "int",
                            "required": False,
                            "choices": [1, 2, 3]
                        }
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [{
                    "something": {
                        "choices": ['The "choices" field can only be set for an "enum" type property.']
                    }
                }]
            }
        })

    def test_create_version_invalid_user_configuration_missing_key(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_integer": {"type": "int", "required": True, "default": 1}
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {
                "configuration": {
                    "user_configuration": [{
                        "demo_integer": {
                            "title": ["This field is required."]
                        }
                    }]
                }
            }
        )

    def test_create_version_invalid_user_configuration_invalid_key(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "demo_integer": {
                            "title": "an integer",
                            "type": "int",
                            "required": True,
                            "default": 1,
                            "some_key": "oh no",
                        }
                    }
                },
                "gpu_usage": "disabled",
            },
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "configuration": {
                "user_configuration": [
                    {
                        "demo_integer": {
                            "some_key": ["Configurable properties can only be defined using the following keys: title, type, required, default, subtype, choices."]
                        }
                    }
                ]
            }
        })

    def test_create_version_invalid_user_configuration_default_value(self):
        cases = [
            ({"type": "int", "default": False}, "int"),
            ({"type": "int", "default": True}, "int"),
            ({"type": "float", "default": False}, "float"),
            ({"type": "float", "default": True}, "float"),
            ({"type": "bool", "default": 0}, "bool"),
            ({"type": "bool", "default": 1}, "bool"),
            ({"type": "string", "default": 1}, "string"),
            ({"type": "dict", "default": ["a", "b"]}, "dict"),
            ({"type": "model", "default": "gigi hadid"}, "model"),
            ({"type": "model", "default": False}, "model")
        ]
        for params, expected in cases:
            with self.subTest(**params):
                response = self.client.post(
                    reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
                    data={
                        "revision_id": str(self.rev2.id),
                        "configuration": {
                            "user_configuration": {
                                "param": {"title": "param", **params}
                            }
                        },
                    },
                    format="json",
                    HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertEqual(response.json(), {"configuration": {"user_configuration": [{"param": {"default": [f"This is not a valid value for a field of type {expected}."]}}]}})

    def test_create_version_user_configuration_model_default_doesnt_exist(self):
        response = self.client.post(
            reverse("api:worker-versions", kwargs={"pk": str(self.worker_dla.id)}),
            data={
                "revision_id": str(self.rev2.id),
                "configuration": {
                    "user_configuration": {
                        "param": {"title": "Model to train", "type": "model", "default": "12341234-1234-1234-1234-123412341234"}
                    }
                },
            },
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"configuration": {"user_configuration": [{"param": {"default": [
            "Model 12341234-1234-1234-1234-123412341234 not found."
        ]}}]}})

    def test_retrieve_version_invalid_id(self):
        self.client.force_login(self.user)
        response = self.client.get(
            reverse("api:version-retrieve", kwargs={"pk": "12341234-1234-1234-1234-123412341234"})
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_retrieve_version_no_login(self):
        """
        Worker versions may be retrieved with no authentication in order
        to see the version on public resources
        """
        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        revision = data.pop("revision")
        self.assertEqual(revision["id"], str(self.rev.id))
        self.assertDictEqual(data, {
            "id": str(self.version_1.id),
            "configuration": {"test": 42},
            "docker_image_iid": self.version_1.docker_image_iid,
            "state": "available",
            "gpu_usage": "disabled",
            "model_usage": FeatureUsage.Disabled.value,
            "worker": {
                "id": str(self.worker_reco.id),
                "name": self.worker_reco.name,
                "type": self.worker_reco.type.slug,
                "slug": self.worker_reco.slug,
            },
            "version": None,
            "created": "2020-02-02T01:23:45.678000Z",
        })

    def test_retrieve_version(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        revision = data.pop("revision")
        self.assertEqual(revision["id"], str(self.rev.id))
        self.assertDictEqual(data, {
            "id": str(self.version_1.id),
            "configuration": {"test": 42},
            "docker_image_iid": self.version_1.docker_image_iid,
            "state": "available",
            "gpu_usage": "disabled",
            "model_usage": FeatureUsage.Disabled.value,
            "worker": {
                "id": str(self.worker_reco.id),
                "name": self.worker_reco.name,
                "type": self.worker_reco.type.slug,
                "slug": self.worker_reco.slug,
            },
            "version": None,
            "created": "2020-02-02T01:23:45.678000Z",
        })

    def test_update_version_requires_login(self):
        response = self.client.patch(
            reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}),
            data={
                "configuration": {"test": "test2"},
                "docker_image_iid": "registry.nerv.co.jp/cruel/angel/thesis:latest",
                "state": "error",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
            }, format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_update_version_invalid_id(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse("api:version-retrieve", kwargs={"pk": "12341234-1234-1234-1234-123412341234"}),
            data={
                "configuration": {"test": "test2"},
                "docker_image_iid": "registry.nerv.co.jp/cruel/angel/thesis:latest",
                "state": "error",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
            }, format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_update_worker_version_non_task(self):
        """
        Only Ponos tasks are allowed to update a worker version
        """
        docker_image_iid = self.version_1.docker_image_iid
        self.version_1.state = WorkerVersionState.Created
        self.version_1.docker_image_iid = None
        self.version_1.save()
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}),
            data={
                "configuration": {"test": "test2"},
                "docker_image_iid": str(docker_image_iid),
                "state": "error",
                "gpu_usage": "disabled",
                "model_usage": FeatureUsage.Disabled.value,
            }, format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {"detail": "Only Ponos tasks are allowed to update a worker version."})

    def test_update_worker_version(self):
        """
        Update worker version artifact, configuration and state
        """
        self.version_1.state = WorkerVersionState.Created
        self.version_1.save()

        response = self.client.patch(
            reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}),
            data={
                "configuration": {"test": "test2"},
                "docker_image_iid": "eva:unit-01",
                "state": "error",
            },
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["id"], str(self.version_1.id))
        self.version_1.refresh_from_db()
        self.assertEqual(data["configuration"], {"test": "test2"})
        self.assertEqual(data["docker_image_iid"], "eva:unit-01")
        self.assertEqual(data["state"], "error")
        self.assertEqual(data["gpu_usage"], "disabled")

    def test_cannot_update_worker_version_revision_ignored(self):
        self.version_1.state = WorkerVersionState.Created
        self.version_1.save()
        with self.assertNumQueries(7):
            response = self.client.patch(
                reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}),
                data={
                    "revision_id": str(self.rev2.id)
                },
                format="json",
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            # revision_id just gets ignored
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["id"], str(self.version_1.id))
        self.version_1.refresh_from_db()
        self.assertEqual(data["revision"]["id"], str(self.rev.id))

    def test_update_version_valid(self):
        self.version_1.state = WorkerVersionState.Created
        self.version_1.save()

        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            farm=self.farm,
        )
        process.run()

        response = self.client.patch(
            reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}),
            data={
                "configuration": {"test": "test2"},
                "docker_image_iid": "registry.nerv.co.jp/cruel/angel/thesis:latest",
                "state": "available",
            },
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["id"], str(self.version_1.id))
        self.assertEqual(data["configuration"], {"test": "test2"})
        self.assertEqual(data["docker_image_iid"], "registry.nerv.co.jp/cruel/angel/thesis:latest")
        self.assertEqual(data["state"], "available")
        self.assertEqual(data["gpu_usage"], "disabled")

    def test_update_version_available_requires_docker_image(self):
        self.version_1.state = WorkerVersionState.Created
        self.version_1.docker_image_iid = None
        self.version_1.save()

        response = self.client.patch(
            reverse("api:version-retrieve", kwargs={"pk": str(self.version_1.id)}),
            data={"state": "available"},
            format="json",
            HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "state": [
                "A worker version must have a `docker_image_iid` set when updated to available."
            ]
        })

    def test_property_slug_explicit(self):
        version_2 = WorkerVersion.objects.create(
            id="01234567-89ab-cdef-fedc-ba9876543210",
            worker=self.worker_reco,
            revision=self.rev2,
            configuration={"test": "test1"},
            gpu_usage=FeatureUsage.Disabled
        )
        self.assertEqual(version_2.slug, "reco_012345")

    def test_property_docker_command_empty(self):
        self.assertEqual(self.version_1.docker_command, None)

    def test_property_docker_command_not_empty(self):
        self.version_1.configuration = {"docker": {"command": "mysupercommand"}}
        self.assertEqual(self.version_1.docker_command, "mysupercommand")
        self.version_1.configuration = {"test": "test1"}

    def test_property_docker_shm_size(self):
        self.version_1.configuration = {"docker": {"shm_size": "32b"}}
        self.assertEqual(self.version_1.docker_shm_size, "32b")

    def test_corpus_worker_version_no_login_public(self):
        self.assertTrue(self.corpus.public)
        # Use multiple versions to ensure we don't get duplicated queries
        corpus_worker_version_1 = self.corpus.worker_version_cache.create(worker_version=self.version_1)
        corpus_worker_version_2 = self.corpus.worker_version_cache.create(worker_version=self.version_2)

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:corpus-versions", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 2,
            "number": 1,
            "previous": None,
            "next": None,
            "results": [{
                "id": str(corpus_worker_version_2.id),
                "worker_configuration": None,
                "model_version": None,
                "worker_version": {
                    "id": str(self.version_2.id),
                    "configuration": {"test": 42},
                    "revision": {
                        "id": str(self.version_2.revision_id),
                        "hash": "1337",
                        "author": "Test user",
                        "message": "My w0rk3r",
                        "created": "2020-02-02T01:23:45.678000Z",
                        "commit_url": "http://my_repo.fake/workers/worker/commit/1337",
                        "refs": []
                    },
                    "gpu_usage": "disabled",
                    "model_usage": FeatureUsage.Disabled.value,
                    "docker_image_iid": self.version_2.docker_image_iid,
                    "state": "available",
                    "worker": {
                        "id": str(self.worker_dla.id),
                        "name": "Document layout analyser",
                        "type": "dla",
                        "slug": "dla",
                    },
                    "created": "2020-02-02T01:23:45.678000Z",
                    "version": None,
                },
            }, {
                "id": str(corpus_worker_version_1.id),
                "worker_configuration": None,
                "model_version": None,
                "worker_version": {
                    "id": str(self.version_1.id),
                    "configuration": {"test": 42},
                    "revision": {
                        "id": str(self.version_1.revision_id),
                        "hash": "1337",
                        "author": "Test user",
                        "message": "My w0rk3r",
                        "created": "2020-02-02T01:23:45.678000Z",
                        "commit_url": "http://my_repo.fake/workers/worker/commit/1337",
                        "refs": []
                    },
                    "gpu_usage": "disabled",
                    "model_usage": FeatureUsage.Disabled.value,
                    "docker_image_iid": self.version_1.docker_image_iid,
                    "state": "available",
                    "worker": {
                        "id": str(self.worker_reco.id),
                        "name": "Recognizer",
                        "type": "recognizer",
                        "slug": "reco",
                    },
                    "created": "2020-02-02T01:23:45.678000Z",
                    "version": None,
                },
            }]
        })

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.filter(public=True))
    def test_corpus_worker_version_no_login_private(self, filter_rights_mock):
        self.corpus.public = False
        self.corpus.save()
        self.corpus.worker_versions.set([self.version_1, self.version_2])

        with self.assertNumQueries(1):
            response = self.client.get(reverse("api:corpus-versions", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertDictEqual(response.json(), {"detail": "No Corpus matches the given query."})

    def test_corpus_worker_version_not_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        # Use multiple versions to ensure we don't get duplicated queries
        corpus_worker_version_1 = self.corpus.worker_version_cache.create(worker_version=self.version_1)
        corpus_worker_version_2 = self.corpus.worker_version_cache.create(worker_version=self.version_2)

        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:corpus-versions", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 2,
            "number": 1,
            "previous": None,
            "next": None,
            "results": [{
                "id": str(corpus_worker_version_2.id),
                "worker_configuration": None,
                "model_version": None,
                "worker_version": {
                    "id": str(self.version_2.id),
                    "configuration": {"test": 42},
                    "revision": {
                        "id": str(self.version_2.revision_id),
                        "hash": "1337",
                        "author": "Test user",
                        "message": "My w0rk3r",
                        "created": "2020-02-02T01:23:45.678000Z",
                        "commit_url": "http://my_repo.fake/workers/worker/commit/1337",
                        "refs": []
                    },
                    "gpu_usage": "disabled",
                    "model_usage": FeatureUsage.Disabled.value,
                    "docker_image_iid": self.version_2.docker_image_iid,
                    "state": "available",
                    "worker": {
                        "id": str(self.worker_dla.id),
                        "name": "Document layout analyser",
                        "type": "dla",
                        "slug": "dla",
                    },
                    "created": "2020-02-02T01:23:45.678000Z",
                    "version": None,
                },
            }, {
                "id": str(corpus_worker_version_1.id),
                "worker_configuration": None,
                "model_version": None,
                "worker_version": {
                    "id": str(self.version_1.id),
                    "configuration": {"test": 42},
                    "revision": {
                        "id": str(self.version_1.revision_id),
                        "hash": "1337",
                        "author": "Test user",
                        "message": "My w0rk3r",
                        "created": "2020-02-02T01:23:45.678000Z",
                        "commit_url": "http://my_repo.fake/workers/worker/commit/1337",
                        "refs": []
                    },
                    "gpu_usage": "disabled",
                    "model_usage": FeatureUsage.Disabled.value,
                    "docker_image_iid": self.version_2.docker_image_iid,
                    "state": "available",
                    "worker": {
                        "id": str(self.worker_reco.id),
                        "name": "Recognizer",
                        "type": "recognizer",
                        "slug": "reco",
                    },
                    "created": "2020-02-02T01:23:45.678000Z",
                    "version": None,
                },
            }]
        })

    def test_corpus_worker_version_list(self):
        self.client.force_login(self.user)
        # Use multiple versions with a model and worker configuration to ensure we don't get duplicated queries
        conf_1 = WorkerConfiguration.objects.create(name="test", configuration={"num": 1337}, worker=self.worker_reco)
        corpus_worker_version_1 = CorpusWorkerVersion.objects.create(
            corpus=self.corpus,
            worker_version=self.version_1,
            model_version=self.model_version,
            worker_configuration=conf_1,
        )
        conf_2 = WorkerConfiguration.objects.create(name="test 2", configuration={"value": "leet"}, worker=self.worker_dla)
        corpus_worker_version_2 = CorpusWorkerVersion.objects.create(
            corpus=self.corpus,
            worker_version=self.version_2,
            model_version=self.model_version,
            worker_configuration=conf_2,
        )

        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:corpus-versions", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 2,
            "number": 1,
            "previous": None,
            "next": None,
            "results": [{
                "id": str(corpus_worker_version_2.id),
                "worker_version": {
                    "id": str(self.version_2.id),
                    "configuration": {"test": 42},
                    "revision": {
                        "id": str(self.version_2.revision_id),
                        "hash": "1337",
                        "author": "Test user",
                        "message": "My w0rk3r",
                        "created": "2020-02-02T01:23:45.678000Z",
                        "commit_url": "http://my_repo.fake/workers/worker/commit/1337",
                        "refs": []
                    },
                    "gpu_usage": "disabled",
                    "model_usage": FeatureUsage.Disabled.value,
                    "docker_image_iid": self.version_2.docker_image_iid,
                    "state": "available",
                    "worker": {
                        "id": str(self.worker_dla.id),
                        "name": "Document layout analyser",
                        "type": "dla",
                        "slug": "dla",
                    },
                    "created": "2020-02-02T01:23:45.678000Z",
                    "version": None,
                },
                "worker_configuration": {
                    "id": str(conf_2.id),
                    "archived": False,
                    "name": "test 2",
                    "configuration": {"value": "leet"},
                },
                "model_version": {
                    "id": str(self.model_version.id),
                    "configuration": {},
                    "model": {
                        "id": str(self.model.id),
                        "name": "Generic model"
                    },
                    "size": 1337,
                    "state": ModelVersionState.Available.value,
                    "tag": None,
                },
            }, {
                "id": str(corpus_worker_version_1.id),
                "worker_version": {
                    "id": str(self.version_1.id),
                    "configuration": {"test": 42},
                    "revision": {
                        "id": str(self.version_1.revision_id),
                        "hash": "1337",
                        "author": "Test user",
                        "message": "My w0rk3r",
                        "created": "2020-02-02T01:23:45.678000Z",
                        "commit_url": "http://my_repo.fake/workers/worker/commit/1337",
                        "refs": []
                    },
                    "gpu_usage": "disabled",
                    "model_usage": FeatureUsage.Disabled.value,
                    "docker_image_iid": self.version_1.docker_image_iid,
                    "state": "available",
                    "worker": {
                        "id": str(self.worker_reco.id),
                        "name": "Recognizer",
                        "type": "recognizer",
                        "slug": "reco",
                    },
                    "created": "2020-02-02T01:23:45.678000Z",
                    "version": None,
                },
                "worker_configuration": {
                    "id": str(conf_1.id),
                    "archived": False,
                    "name": "test",
                    "configuration": {"num": 1337},
                },
                "model_version": {
                    "id": str(self.model_version.id),
                    "configuration": {},
                    "model": {
                        "id": str(self.model.id),
                        "name": "Generic model"
                    },
                    "size": 1337,
                    "state": ModelVersionState.Available.value,
                    "tag": None,
                },
            }],
        })

    def test_corpus_worker_version_no_right(self):
        self.user.rights.all().delete()
        self.client.force_login(self.user)
        corpus_worker_version = self.corpus.worker_version_cache.create(worker_version=self.version_1)

        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:corpus-versions", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "number": 1,
            "previous": None,
            "next": None,
            "results": [{
                "id": str(corpus_worker_version.id),
                "worker_configuration": None,
                "model_version": None,
                "worker_version": {
                    "id": str(self.version_1.id),
                    "configuration": {"test": 42},
                    "revision": {
                        "id": str(self.version_1.revision_id),
                        "hash": "1337",
                        "author": "Test user",
                        "message": "My w0rk3r",
                        "created": "2020-02-02T01:23:45.678000Z",
                        "commit_url": "http://my_repo.fake/workers/worker/commit/1337",
                        "refs": []
                    },
                    "gpu_usage": "disabled",
                    "model_usage": FeatureUsage.Disabled.value,
                    "docker_image_iid": self.version_1.docker_image_iid,
                    "state": "available",
                    "worker": {
                        "id": str(self.worker_reco.id),
                        "name": "Recognizer",
                        "type": "recognizer",
                        "slug": "reco",
                    },
                    "created": "2020-02-02T01:23:45.678000Z",
                    "version": None,
                }
            }]
        })
