from datetime import datetime, timezone
from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.process.models import Repository, Worker
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Model
from arkindex.users.models import Role


class TestModelCompatibleWorkerManage(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.worker1 = Worker.objects.get(slug="generic")
        cls.worker2 = Worker.objects.get(slug="dla")

        cls.worker2.memberships.create(user=cls.user, level=Role.Contributor.value)

        cls.model1 = Model.objects.create(name="A model")
        cls.model2 = Model.objects.create(name="A public model", public=True)

        cls.model1.compatible_workers.set([cls.worker1])
        cls.model2.compatible_workers.set([cls.worker2])

        cls.model1.memberships.create(user=cls.user, level=Role.Contributor.value)

    def test_create_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.model1.id),
                    "worker": str(self.worker2.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {
            "detail": "Authentication credentials were not provided.",
        })

    def test_create_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.model1.id),
                    "worker": str(self.worker2.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {
            "detail": "You do not have permission to perform this action.",
        })

    def test_create_model_not_found(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.worker1.id),
                    "worker": str(self.worker2.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model": [f'Invalid pk "{self.worker1.id}" - object does not exist.'],
        })

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_create_requires_model_contributor(self, filter_rights_mock):
        self.model1.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)

        filter_rights_mock.side_effect = [
            Model.objects.none(),
            Worker.objects.all(),
            Repository.objects.all(),
        ]

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.model1.id),
                    "worker": str(self.worker2.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model": [f'Invalid pk "{self.model1.id}" - object does not exist.'],
        })

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Model, Role.Contributor.value),
            call(self.user, Worker, Role.Contributor.value),
            call(self.user, Repository, Role.Contributor.value),
        ])

    def test_create_worker_not_found(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.model1.id),
                    "worker": str(self.model1.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "worker": [f'Invalid pk "{self.model1.id}" - object does not exist.'],
        })

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_create_requires_worker_contributor(self, filter_rights_mock):
        self.worker2.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)

        filter_rights_mock.side_effect = [
            Model.objects.all(),
            Worker.objects.none(),
            Repository.objects.none(),
        ]

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.model1.id),
                    "worker": str(self.worker2.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "worker": [f'Invalid pk "{self.worker2.id}" - object does not exist.'],
        })

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Model, Role.Contributor.value),
            call(self.user, Worker, Role.Contributor.value),
            call(self.user, Repository, Role.Contributor.value),
        ])

    def test_create_nothing_found(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.worker2.id),
                    "worker": str(self.model1.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model": [f'Invalid pk "{self.worker2.id}" - object does not exist.'],
            "worker": [f'Invalid pk "{self.model1.id}" - object does not exist.'],
        })

    def test_create(self):
        self.client.force_login(self.user)

        # 2 queries to retrieve the user and session
        # 2 queries to retrieve the model and worker
        # 1 query to check that the relationship does not already exist
        # 1 query to insert
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.model1.id),
                    "worker": str(self.worker2.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(response.json(), {
            "model": str(self.model1.id),
            "worker": str(self.worker2.id),
        })
        self.assertQuerySetEqual(
            self.model1.compatible_workers.all(),
            [self.worker1, self.worker2],
            ordered=False,
        )
        self.assertQuerySetEqual(
            self.model2.compatible_workers.all(),
            [self.worker2],
            ordered=False,
        )

    def test_create_worker_public(self):
        # User has no access to worker1, but will be able to use it because it is public
        self.worker1.public = True
        self.worker1.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.model1.id),
                    "worker": str(self.worker2.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(response.json(), {
            "model": str(self.model1.id),
            "worker": str(self.worker2.id),
        })
        self.assertQuerySetEqual(
            self.model1.compatible_workers.all(),
            [self.worker1, self.worker2],
            ordered=False,
        )
        self.assertQuerySetEqual(
            self.model2.compatible_workers.all(),
            [self.worker2],
            ordered=False,
        )

    def test_create_admin(self):
        self.client.force_login(self.superuser)

        # No content types retrieved because all ACL checks are ignored
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.model1.id),
                    "worker": str(self.worker2.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(response.json(), {
            "model": str(self.model1.id),
            "worker": str(self.worker2.id),
        })
        self.assertQuerySetEqual(
            self.model1.compatible_workers.all(),
            [self.worker1, self.worker2],
            ordered=False,
        )
        self.assertQuerySetEqual(
            self.model2.compatible_workers.all(),
            [self.worker2],
            ordered=False,
        )

    def test_create_exists(self):
        self.model1.compatible_workers.add(self.worker2)
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.model1.id),
                    "worker": str(self.worker2.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "worker": ["This worker is already marked as compatible with this model."],
        })
        self.assertQuerySetEqual(
            self.model1.compatible_workers.all(),
            [self.worker1, self.worker2],
            ordered=False,
        )
        self.assertQuerySetEqual(
            self.model2.compatible_workers.all(),
            [self.worker2],
            ordered=False,
        )

    def test_create_archived(self):
        self.model1.archived = datetime.now(timezone.utc)
        self.model1.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.model1.id),
                    "worker": str(self.worker2.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model": ["This model is archived."],
        })
        self.assertQuerySetEqual(
            self.model1.compatible_workers.all(),
            [self.worker1],
            ordered=False,
        )
        self.assertQuerySetEqual(
            self.model2.compatible_workers.all(),
            [self.worker2],
            ordered=False,
        )

    def test_destroy_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.delete(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.model1.id),
                    "worker": str(self.worker1.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {
            "detail": "Authentication credentials were not provided.",
        })

    def test_destroy_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.delete(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.model1.id),
                    "worker": str(self.worker1.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {
            "detail": "You do not have permission to perform this action.",
        })

    def test_destroy_model_not_found(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.worker2.id),
                    "worker": str(self.worker2.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model": [f'Invalid pk "{self.worker2.id}" - object does not exist.'],
        })

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_destroy_requires_model_contributor(self, filter_rights_mock):
        self.model1.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)

        filter_rights_mock.side_effect = [
            Model.objects.none(),
            Worker.objects.all(),
            Repository.objects.all(),
        ]

        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.model1.id),
                    "worker": str(self.worker2.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model": [f'Invalid pk "{self.model1.id}" - object does not exist.'],
        })

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Model, Role.Contributor.value),
            call(self.user, Worker, Role.Contributor.value),
            call(self.user, Repository, Role.Contributor.value),
        ])

    def test_destroy_worker_not_found(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.model1.id),
                    "worker": str(self.model1.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "worker": [f'Invalid pk "{self.model1.id}" - object does not exist.'],
        })

    @patch("arkindex.users.managers.BaseACLManager.filter_rights")
    def test_destroy_requires_worker_contributor(self, filter_rights_mock):
        self.worker2.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)

        filter_rights_mock.side_effect = [
            Model.objects.all(),
            Worker.objects.none(),
            Repository.objects.none(),
        ]

        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.model1.id),
                    "worker": str(self.worker2.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "worker": [f'Invalid pk "{self.worker2.id}" - object does not exist.'],
        })

        self.assertListEqual(filter_rights_mock.call_args_list, [
            call(self.user, Model, Role.Contributor.value),
            call(self.user, Worker, Role.Contributor.value),
            call(self.user, Repository, Role.Contributor.value),
        ])

    def test_destroy_nothing_found(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.worker2.id),
                    "worker": str(self.model1.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model": [f'Invalid pk "{self.worker2.id}" - object does not exist.'],
            "worker": [f'Invalid pk "{self.model1.id}" - object does not exist.'],
        })

    def test_destroy(self):
        self.client.force_login(self.user)
        # Have two workers to check not every worker will be removed
        self.model1.compatible_workers.set([self.worker1, self.worker2])

        # 2 queries to retrieve the user and session
        # 2 queries to retrieve the model and worker
        # 1 query to retrieve the existing relationship
        # 1 query to delete
        with self.assertNumQueries(6):
            response = self.client.delete(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.model1.id),
                    "worker": str(self.worker2.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertQuerySetEqual(
            self.model1.compatible_workers.all(),
            [self.worker1],
            ordered=False,
        )
        self.assertQuerySetEqual(
            self.model2.compatible_workers.all(),
            [self.worker2],
            ordered=False,
        )

    def test_destroy_worker_public(self):
        # User has no access to worker1, but will be able to use it because it is public
        self.worker1.public = True
        self.worker1.save()
        self.model1.compatible_workers.add(self.worker2)
        self.client.force_login(self.user)

        with self.assertNumQueries(6):
            response = self.client.delete(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.model1.id),
                    "worker": str(self.worker1.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertQuerySetEqual(
            self.model1.compatible_workers.all(),
            [self.worker2],
            ordered=False,
        )
        self.assertQuerySetEqual(
            self.model2.compatible_workers.all(),
            [self.worker2],
            ordered=False,
        )

    def test_destroy_admin(self):
        self.model1.compatible_workers.add(self.worker2)
        self.client.force_login(self.superuser)

        # No content types retrieved because all ACL checks are ignored
        with self.assertNumQueries(6):
            response = self.client.delete(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.model1.id),
                    "worker": str(self.worker1.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertQuerySetEqual(
            self.model1.compatible_workers.all(),
            [self.worker2],
            ordered=False,
        )
        self.assertQuerySetEqual(
            self.model2.compatible_workers.all(),
            [self.worker2],
            ordered=False,
        )

    def test_destroy_does_not_exist(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.delete(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.model1.id),
                    "worker": str(self.worker2.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(response.json(), {
            "detail": "No Model_compatible_workers matches the given query.",
        })
        self.assertQuerySetEqual(
            self.model1.compatible_workers.all(),
            [self.worker1],
            ordered=False,
        )
        self.assertQuerySetEqual(
            self.model2.compatible_workers.all(),
            [self.worker2],
            ordered=False,
        )

    def test_destroy_archived(self):
        self.model1.archived = datetime.now(timezone.utc)
        self.model1.save()
        self.model1.compatible_workers.add(self.worker2)
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:model-compatible-worker-manage", kwargs={
                    "model": str(self.model1.id),
                    "worker": str(self.worker2.id),
                })
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "model": ["This model is archived."],
        })
        self.assertQuerySetEqual(
            self.model1.compatible_workers.all(),
            [self.worker1, self.worker2],
            ordered=False,
        )
        self.assertQuerySetEqual(
            self.model2.compatible_workers.all(),
            [self.worker2],
            ordered=False,
        )
